package orm.structure;

import java.io.*;
import java.lang.reflect.*;
import java.util.*;
import orm.api.*;

import orm.generators.*;
import orm.runtime.ORMConfiguration;
import orm.api.relations.*;

public class ClassStructure {

	
	private final String implDir = ORMConfiguration.getPropertyValue("ImplementationDir");
	private final String genDir = ORMConfiguration.getPropertyValue("OrmGenerationDir"); 
	private final String implLibPath = ORMConfiguration.getPropertyValue("ImplementationLibPath");
	
	private DataStructure dataStructure;
	private Hashtable<Class, String> classes = new Hashtable<Class, String>();
	
	protected ClassStructure(DataStructure dataStructure) {
		this.dataStructure = dataStructure;
	}
	
	protected void init (StartUpType type) throws ORMException {

		if (! new File(genDir + "/Impl.jar").exists()) {
			type = StartUpType.Create;
		}
		if (StartUpType.Create.equals(type) || StartUpType.Update.equals(type)) {
			
			clearDirectoryContent(genDir);
			File d = new File(genDir);
			d.mkdirs();
			
			for(Class c : dataStructure.getClasses()) {
				generateSourceFile(c);
			}
			compileSource();
			//copyToImplDir();
		}
	}
	
	public static ClassStructure getStructure(DataStructure dataStructure, StartUpType type) throws ORMException {
		ClassStructure res = new ClassStructure(dataStructure);
		res.init(type);
		return res;
	}
	
	private void clearDirectoryContent(String dir) {
		File d = new File(dir);
		if (!d.exists()) return;
		if (!d.isDirectory()) return;
		clearContent(d);
	}
	
	private void clearContent(File dir) {
		for(File f : dir.listFiles()) {
			if (f.isDirectory()) {
				clearContent(f);
			} else {
				f.delete();
			}
		}
		dir.delete();
	}
	
	
	
	protected String getSourceFile(Class c) {
		return genDir + "/orm/impl/" + c.getSimpleName() + ".java";
	}
	protected String getImplClassName(Class c) {
		return "orm.impl." + c.getName();
	}
	public String generateSourceFile(Class c) throws ORMException {
		String fileName = getSourceFile(c);
		File source = new File(fileName);
		if (!source.exists()) {
			try {
				if (!source.getParentFile().exists()) {
					source.getParentFile().mkdirs();
				}
				source.createNewFile();
			} catch (IOException ioe) {
				throw new ORMException(ioe);
			}
		}
		if (!source.exists()){
			throw new ORMException("Could not create " + source.getAbsolutePath() + " file");
		}
		 
		
		
		BufferedGenWriter bw = null;
		
		try {
			bw = new BufferedGenWriter(new FileWriter(source));
		} catch (IOException e) {
			throw new ORMException(e);
		}
		
		try {
			
			String newClassName = c.getSimpleName();
			bw.writeLine("package orm.impl;");
			classes.put(c, "orm.impl." + newClassName);
			bw.newLine();
			//insert import statements
			bw.writeLine("import orm.structure.*;");
			bw.writeLine("import orm.api.*;");
			bw.writeLine("import orm.runtime.*;");
			bw.writeLine("import java.lang.reflect.*;");
			bw.newLine();
			//insert import statements end
			bw.writeLine("public class " + newClassName + " extends " + c.getName() + " implements IOrmPersistent" );
			bw.writeLine("{");
			//body beginning
			bw.newLine();
			bw.writeLine("private IOrmEntity _entity = null;",1);
			bw.writeLine("private boolean _isChanged = false;",1);
			bw.newLine();
			//constructores
			int consNum = 0;
			for(Constructor cons : c.getDeclaredConstructors()) {
				String modifier = getModifier(cons);
				if (modifier.length() == 0) continue;
				
				bw.write(modifier + " " + newClassName + "(",1);
				bw.write("IOrmEntity ormEntity");
				int paramIndex = 0;
				for(Class param : cons.getParameterTypes()) {
					bw.write(",");
					bw.write(param.getCanonicalName());
					bw.write(" arg" + paramIndex);
					paramIndex++;
				}
				bw.writeLine(") throws ORMException");
				bw.writeLine("{",1);
				bw.write("super(",2);
				int paramLen = cons.getParameterTypes().length;
				for(paramIndex = 0; paramIndex < paramLen; paramIndex++ ) {
					if (paramIndex > 0) {
						bw.write(",");
					}
					bw.write("arg" + paramIndex);
					paramIndex++;
				}
				bw.writeLine(");");
				bw.writeLine("init(ormEntity);",2);
				bw.writeLine("}",1);
				consNum++;
			}//constructors end
			if (consNum == 0) {
				bw.writeLine("public " + newClassName + "(IOrmEntity ormEntity)",1);
				bw.writeLine("{", 1);
				bw.writeLine("init(ormEntity);", 2);
				bw.writeLine("}", 1);
			}
			//init here
			bw.writeLine("private void init(IOrmEntity ormEntity) throws ORMException",1);
			bw.writeLine("{",1);
			bw.writeLine("_entity = ormEntity;", 2);
			bw.writeLine("",2);
			bw.writeLine("//initialize the fields", 2);
			for(IFieldStructure f : this.dataStructure.getEntityStructure(c).getFields()) {
				int modifiers = f.getField().getModifiers();
				boolean isPrivate = !Modifier.isPublic(modifiers) && ! Modifier.isProtected(modifiers);
				if (isPrivate) {
					bw.writeLine("try {", 2);
					bw.writeLine("Field _f_" + f.getField().getName() + " = getClass().getSuperclass().getField(\"" + f.getField().getName() + "\");", 3);
					bw.writeLine("_f_" + f.getField().getName() + ".setAccessible(true);", 3);
					bw.writeLine("_f_" + f.getField().getName() + ".set(this, _entity.getFieldData(\"" + f.getFieldName() + "\"));", 3);
					bw.writeLine("} catch (Exception e) { throw new ORMException(e);}", 2);
				} else {
					bw.writeLine(f.getField().getName() + " = (" + getTypeForCasting(f.getField()) + ") _entity.getFieldData(\"" + f.getFieldName() + "\");", 2);
				}
			}
			for(IRelation r : this.dataStructure.getEntityStructure(c).getRelations()){
				IFieldStructure f = r.getFieldStructure();
				bw.writeLine(f.getField().getName() + " = (" + getTypeForCasting(f.getField()) + ") _entity.getFieldData(\"" + f.getFieldName() + "\");", 2);
			}
			bw.writeLine("}",1);
			//init end
			//IOrmPersistent start
			bw.writeLine("// implement IOrmPersistent methods",1);
			bw.writeLine("public void commit() throws ORMException {",1);
			bw.writeLine("OrmRuntime.commit(this);",2);
			bw.writeLine("}",1);
			bw.writeLine("public void reset() throws ORMException {",1);
			bw.writeLine("init(_entity);",2);
			bw.writeLine("}",1);
			//getOrmEntityWithUpdatedData()
			bw.writeLine("public IOrmEntity getOrmEntityWithUpdatedData() throws ORMException {",1);
			
			for(IFieldStructure f : this.dataStructure.getEntityStructure(c).getFields()) {
				int modifiers = f.getField().getModifiers();
				boolean isPrivate = !Modifier.isPublic(modifiers) && ! Modifier.isProtected(modifiers);
				if(isPrivate) {
					bw.writeLine("try {", 2);
					bw.writeLine("Field _f_" + f.getField().getName() + " = getClass().getSuperclass().getField(\"" + f.getField().getName() + "\");", 3);
					bw.writeLine("_f_" + f.getField().getName() + ".setAccessible(true);", 3);
					bw.writeLine("_entity.setFieldData(\"" + f.getFieldName() + "\",_f_" + f.getField().getName() + ".get(this));", 3);
					bw.writeLine("} catch (Exception e) { throw new ORMException(e);}", 2);
				} else {
					bw.writeLine("_entity.setFieldData(\"" + f.getFieldName() + "\"," + f.getField().getName() + ");", 2);
				}
			}
			for(IRelation r : this.dataStructure.getEntityStructure(c).getRelations()){
				IFieldStructure f = r.getFieldStructure();
				bw.writeLine("_entity.setFieldData(\"" + f.getFieldName() + "\"," + f.getField().getName() + ");", 2);
			}
			bw.writeLine("return _entity;",2);
			bw.writeLine("}",1);
			//IOrmPersistent end
			
//			body end
			bw.write("}");
			bw.newLine();
			
		} catch (IOException e) {
			throw new ORMException(e);
		} finally {
			try {
				bw.close();
			} catch (IOException e) {
				throw new ORMException(e);
			}
		}
		
		return fileName;
	}
	
	private String getTypeForCasting(Field f) {
		return getTypeForCasting(f.getType(), true);
	}
	private String getTypeForCasting(Class fieldType, boolean boxPrimitives) {
		if (fieldType.isArray()) {
			return getTypeForCasting(fieldType.getComponentType(), false) + "[]";
		}
		if (boxPrimitives && fieldType.isPrimitive()) {
			if (fieldType.equals(int.class)) {
				return "Integer";
			} else if (fieldType.equals(short.class)) {
				return "Short";
			} else if (fieldType.equals(long.class)) {
				return "Long";
			} else if (fieldType.equals(float.class)) {
				return "Float";
			} else if (fieldType.equals(double.class)) {
				return "Double";
			} else if (fieldType.equals(byte.class)) {
				return "Byte";
			} else if (fieldType.equals(boolean.class)) {
				return "Boolean";
			} else if (fieldType.equals(char.class)) {
				return "Character";
			} else return null;
			
		} else {
			return fieldType.getName();
		}
	}
	private void compileSource() throws ORMException {
		JavaCompiler jc = new JavaCompiler();
		boolean isOK = jc.CompileJar(new File(genDir + "/orm/impl"), this.implDir, new File(genDir + "/Impl.jar"));
		if (!isOK) throw new ORMException("Error in compiling jar file :" + jc.getMessage());
	}
	
	private void copyToImplDir() throws ORMException {
		File jar = new File(genDir + "/orm/impl/Impl.jar");
		if (!jar.exists()) {
			throw new ORMException("Compiled jar file:" + genDir + "/orm/impl/Impl.jar does not exists");
		}
		File destFile = new File(implDir + implLibPath + "Impl.jar");

		try {
			FileOutputStream fw = new FileOutputStream(destFile,false);
			try {
				FileInputStream fr = new FileInputStream(jar);
				try {
					byte[] buff = new byte[124];
					int read = 0;
					while ((read = fr.read(buff)) > 0) {
						fw.write(buff, 0, read);
					}
				} finally {
					fr.close();
				}
			} finally {
				fw.close();
			}
		} catch (Exception e) {
			throw new ORMException(e);
		}
		
	}
	private String getModifier(Constructor cons) {
		String res ="";
		int modifiers = cons.getModifiers();
		if (Modifier.isPublic(modifiers)) {
			res = "public";
		} else if (Modifier.isProtected(modifiers)) {
			res = "protected";
		}
		return res;
	}
	
}
